# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common.data_types import guess_numpy_type, guess_proto_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import (
    apply_add, apply_log, apply_mul, apply_identity, apply_normalizer)


def convert_sklearn_tfidf_transformer(scope: Scope, operator: Operator,
                                      container: ModelComponentContainer):
    # TODO: use sparse containers when available
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    float_type = dtype
    # onnx_proto.TensorProto.FLOAT
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    data = operator.input_full_names
    output_name = scope.get_unique_variable_name('tfidftr_output')

    if op.sublinear_tf:
        # code scikit-learn
        # np.log(X.data, X.data) --> does not apply on null coefficient
        # X.data += 1
        # ONNX does not support sparse tensors before opset < 11
        # approximated by X.data += 1 --> np.log(X.data, X.data)
        if operator.target_opset < 11:
            plus1 = scope.get_unique_variable_name("plus1")
            C = operator.inputs[0].type.shape[1]
            ones = scope.get_unique_variable_name("ones")
            cst = np.ones((C,), dtype=float_type)
            container.add_initializer(ones, proto_dtype, [C], cst.flatten())
            apply_add(scope, data + [ones], plus1, container, broadcast=1)
            plus1logged = scope.get_unique_variable_name("plus1logged")
            apply_log(scope, plus1, plus1logged, container)
            data = [plus1logged]
        else:
            # sparse containers have not yet been implemented.
            raise RuntimeError(
                "ONNX does not support sparse tensors before opset < 11, "
                "sublinear_tf must be False.")

    if op.use_idf:
        cst = op.idf_.astype(float_type)
        if len(cst.shape) > 1:
            cst = np.diag(cst)
        cst = cst.ravel().flatten()
        shape = [len(cst)]
        idfcst = scope.get_unique_variable_name('idfcst')
        container.add_initializer(idfcst, proto_dtype, shape, cst)
        apply_mul(scope, data + [idfcst], output_name, container, broadcast=1)
    else:
        output_name = data[0]

    if op.norm is not None:
        norm_name = scope.get_unique_variable_name('tfidftr_norm')
        apply_normalizer(
            scope, output_name, norm_name, container,
            norm=op.norm.upper(), use_float=float_type == np.float32)
        output_name = norm_name

    options = container.get_options(op, dict(nan=False))
    replace_by_nan = options.get('nan', False)
    if replace_by_nan:
        # This part replaces all null values by nan.
        cst_nan_name = scope.get_unique_variable_name('nan_name')
        container.add_initializer(cst_nan_name, proto_dtype, [1], [np.nan])
        cst_zero_name = scope.get_unique_variable_name('zero_name')
        container.add_initializer(cst_zero_name, proto_dtype, [1], [0])

        mask_name = scope.get_unique_variable_name('mask_name')
        container.add_node('Equal', [output_name, cst_zero_name],
                           mask_name,
                           name=scope.get_unique_operator_name('Equal'))

        where_name = scope.get_unique_variable_name('where_name')
        container.add_node('Where', [mask_name, cst_nan_name, output_name],
                           where_name,
                           name=scope.get_unique_operator_name('Where'))
        output_name = where_name

    apply_identity(scope, output_name, operator.output_full_names, container)


register_converter('SklearnTfidfTransformer',
                   convert_sklearn_tfidf_transformer,
                   options={'nan': [True, False]})



# SPDX-License-Identifier: Apache-2.0

from .. import update_registered_converter
from ..shape_calculators.text_vectorizer import (
    calculate_sklearn_text_vectorizer_output_shapes)
from ..operator_converters.text_vectoriser import (
    convert_sklearn_text_vectorizer)
from ..operator_converters.tfidf_vectoriser import (
    convert_sklearn_tfidf_vectoriser)
from .sklearn_text import TraceableCountVectorizer, TraceableTfidfVectorizer


def register():
    """Register converter for TraceableCountVectorizer,
    TraceableTfidfVectorizer."""
    update_registered_converter(
        TraceableCountVectorizer, "Skl2onnxTraceableCountVectorizer",
        calculate_sklearn_text_vectorizer_output_shapes,
        convert_sklearn_text_vectorizer,
        options={'tokenexp': None, 'separators': None,
                 'nan': [True, False],
                 'keep_empty_string': [True, False]})

    update_registered_converter(
        TraceableTfidfVectorizer, "Skl2onnxTraceableTfidfVectorizer",
        calculate_sklearn_text_vectorizer_output_shapes,
        convert_sklearn_tfidf_vectoriser,
        options={'tokenexp': None, 'separators': None,
                 'nan': [True, False],
                 'keep_empty_string': [True, False]})



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes
from ..common.data_types import FloatTensorType, DoubleTensorType
from ..common.utils import check_input_and_output_types


def calculate_sklearn_gaussian_process_regressor_shape(operator):
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType])
    if len(operator.inputs) != 1:
        raise RuntimeError("Only one input vector is allowed for "
                           "GaussianProcessRegressor.")
    if len(operator.outputs) not in (1, 2):
        raise RuntimeError("One output is expected for "
                           "GaussianProcessRegressor.")

    variable = operator.inputs[0]

    N = variable.get_first_dimension()
    op = operator.raw_operator

    # Output 1 is mean
    # Output 2 is cov or std
    if hasattr(op, 'y_train_') and op.y_train_ is not None:
        dim = 1 if len(op.y_train_.shape) == 1 else op.y_train_.shape[1]
    else:
        dim = 1
    operator.outputs[0].type.shape = [N, dim]


register_shape_calculator('SklearnGaussianProcessRegressor',
                          calculate_sklearn_gaussian_process_regressor_shape)
register_shape_calculator('SklearnGaussianProcessClassifier',
                          calculate_linear_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._apply_operation import (
    apply_add, apply_cast, apply_clip, apply_concat, apply_div, apply_exp,
    apply_identity, apply_mul, apply_reciprocal, apply_reshape, apply_sub)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_numpy_type,
    guess_proto_type)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto


def _decision_function(scope, operator, container, model, proto_type):
    """Predict for linear model.
    score = X * coefficient + intercept
    """
    coef_name = scope.get_unique_variable_name('coef')
    intercept_name = scope.get_unique_variable_name('intercept')
    matmul_result_name = scope.get_unique_variable_name(
        'matmul_result')
    score_name = scope.get_unique_variable_name('score')
    coef = model.coef_.T

    container.add_initializer(coef_name, proto_type,
                              coef.shape, coef.ravel())
    container.add_initializer(intercept_name, proto_type,
                              model.intercept_.shape, model.intercept_)

    input_name = operator.inputs[0].full_name
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=proto_type)
        input_name = cast_input_name

    container.add_node(
        'MatMul', [input_name, coef_name],
        matmul_result_name,
        name=scope.get_unique_operator_name('MatMul'))
    apply_add(scope, [matmul_result_name, intercept_name],
              score_name, container, broadcast=0)
    return score_name


def _handle_zeros(scope, container, scores, proba, reduced_proba, num_classes,
                  proto_type):
    """Handle cases where reduced_proba values are zeros to avoid NaNs in
    class probability scores because of divide by 0 when we calculate
    proba / reduced_proba in _normalise_proba().
    This is done by replacing reduced_proba values of 0s with
    num_classes and corresponding proba values with 1.
    """
    num_classes_name = scope.get_unique_variable_name('num_classes')
    bool_reduced_proba_name = scope.get_unique_variable_name(
        'bool_reduced_proba')
    bool_not_reduced_proba_name = scope.get_unique_variable_name(
        'bool_not_reduced_proba')
    not_reduced_proba_name = scope.get_unique_variable_name(
        'not_reduced_proba')
    proba_updated_name = scope.get_unique_variable_name('proba_updated')
    mask_name = scope.get_unique_variable_name('mask')
    reduced_proba_updated_name = scope.get_unique_variable_name(
        'reduced_proba_updated')

    container.add_initializer(num_classes_name, proto_type,
                              [], [num_classes])

    apply_cast(scope, reduced_proba, bool_reduced_proba_name, container,
               to=onnx_proto.TensorProto.BOOL)
    container.add_node('Not', bool_reduced_proba_name,
                       bool_not_reduced_proba_name,
                       name=scope.get_unique_operator_name('Not'))
    apply_cast(scope, bool_not_reduced_proba_name, not_reduced_proba_name,
               container, to=proto_type)
    apply_add(scope, [proba, not_reduced_proba_name],
              proba_updated_name, container, broadcast=1)
    apply_mul(scope, [not_reduced_proba_name, num_classes_name],
              mask_name, container, broadcast=1)
    apply_add(scope, [reduced_proba, mask_name],
              reduced_proba_updated_name, container, broadcast=0)
    return proba_updated_name, reduced_proba_updated_name


def _normalise_proba(scope, operator, container, scores, proba, num_classes,
                     unity_name, proto_type):
    reduced_proba_name = scope.get_unique_variable_name('reduced_proba')
    sub_result_name = scope.get_unique_variable_name('sub_result')

    if num_classes == 2:
        apply_sub(scope, [unity_name, proba],
                  sub_result_name, container, broadcast=1)
        apply_concat(scope, [sub_result_name, proba],
                     operator.outputs[1].full_name, container, axis=1)
    else:
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', proba, reduced_proba_name, axes=[1],
                name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [proba, axis_name], reduced_proba_name,
                name=scope.get_unique_operator_name('ReduceSum'))
        proba_updated, reduced_proba_updated = _handle_zeros(
            scope, container, scores, proba, reduced_proba_name, num_classes,
            proto_type)
        apply_div(scope, [proba_updated, reduced_proba_updated],
                  operator.outputs[1].full_name, container, broadcast=1)
    return operator.outputs[1].full_name


def _predict_proba_log(scope, operator, container, scores, num_classes,
                       proto_type):
    """Probability estimation for SGDClassifier with loss=log and
    Logistic Regression.
    Positive class probabilities are computed as
        1. / (1. + exp(-scores))
        multiclass is handled by normalising that over all classes.
    """
    if num_classes >= 3 or container.target_opset < 13:
        negated_scores_name = scope.get_unique_variable_name('negated_scores')
        negate_name = scope.get_unique_variable_name('negate')
        exp_result_name = scope.get_unique_variable_name('exp_result')
        unity_name = scope.get_unique_variable_name('unity')
        add_result_name = scope.get_unique_variable_name('add_result')
        proba_name = scope.get_unique_variable_name('proba')

        container.add_initializer(negate_name, proto_type, [], [-1])
        container.add_initializer(unity_name, proto_type, [], [1])

        apply_mul(scope, [scores, negate_name],
                  negated_scores_name, container, broadcast=1)
        apply_exp(scope, negated_scores_name, exp_result_name, container)
        apply_add(scope, [exp_result_name, unity_name],
                  add_result_name, container, broadcast=1)
        apply_reciprocal(scope, add_result_name, proba_name, container)
        return _normalise_proba(scope, operator, container, scores, proba_name,
                                num_classes, unity_name, proto_type)

    # Sigmoid cannot be used for num_classes > 2 because
    # onnxruntime has a different implementation than numpy.
    # It introduces discrepancies when x < 1e16.
    # Below that threshold, Sigmoid must be replaced by Exp
    # because Sigmoid is not an increasing function.
    sigmo = scope.get_unique_variable_name('sigmoid')
    container.add_node('Sigmoid', [scores], [sigmo],
                       name=scope.get_unique_operator_name('Sigmoid'))

    unity_name = scope.get_unique_variable_name('unity')
    container.add_initializer(unity_name, proto_type, [1], [1])

    sigmo_0 = scope.get_unique_variable_name('sigmo_0')
    container.add_node('Sub', [unity_name, sigmo], [sigmo_0],
                       name=scope.get_unique_operator_name('Sub'))
    apply_concat(scope, [sigmo_0, sigmo], [operator.outputs[1].full_name],
                 container, axis=1)
    return operator.outputs[1].full_name


def _predict_proba_modified_huber(scope, operator, container,
                                  scores, num_classes, proto_type):
    """Probability estimation for SGDClassifier with
    loss=modified_huber.
    Multiclass probability estimates are derived from binary
    estimates by normalisation.
    Binary probability estimates are given by
    (clip(scores, -1, 1) + 1) / 2.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    unity_name = scope.get_unique_variable_name('unity')
    constant_name = scope.get_unique_variable_name('constant')
    add_result_name = scope.get_unique_variable_name('add_result')
    proba_name = scope.get_unique_variable_name('proba')
    clipped_scores_name = scope.get_unique_variable_name('clipped_scores')

    container.add_initializer(unity_name, proto_type,
                              [], [1])
    container.add_initializer(constant_name, proto_type,
                              [], [2])

    apply_clip(scope, scores, clipped_scores_name, container,
               max=np.array(1, dtype=dtype),
               min=np.array(-1, dtype=dtype))
    apply_add(scope, [clipped_scores_name, unity_name],
              add_result_name, container, broadcast=1)
    apply_div(scope, [add_result_name, constant_name],
              proba_name, container, broadcast=1)
    return _normalise_proba(scope, operator, container, scores, proba_name,
                            num_classes, unity_name, proto_type)


def convert_sklearn_sgd_classifier(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    """Converter for SGDClassifier."""
    sgd_op = operator.raw_operator
    classes = get_label_classes(scope, sgd_op)
    class_type = onnx_proto.TensorProto.STRING
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT

    if np.issubdtype(classes.dtype, np.floating):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    classes_name = scope.get_unique_variable_name('classes')
    predicted_label_name = scope.get_unique_variable_name(
        'predicted_label')
    final_label_name = scope.get_unique_variable_name('final_label')

    container.add_initializer(classes_name, class_type,
                              classes.shape, classes)

    scores = _decision_function(scope, operator, container, sgd_op, proto_type)
    options = container.get_options(sgd_op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']
    if sgd_op.loss == 'log' and not use_raw_scores:
        proba = _predict_proba_log(scope, operator, container, scores,
                                   len(classes), proto_type)
    elif sgd_op.loss == 'modified_huber' and not use_raw_scores:
        proba = _predict_proba_modified_huber(
            scope, operator, container, scores, len(classes),
            proto_type)
    else:
        if len(classes) == 2:
            negate_name = scope.get_unique_variable_name('negate')
            negated_scores_name = scope.get_unique_variable_name(
                'negated_scores')

            container.add_initializer(
                negate_name, proto_type, [], [-1])

            apply_mul(scope, [scores, negate_name],
                      negated_scores_name, container, broadcast=1)
            apply_concat(scope, [negated_scores_name, scores],
                         operator.outputs[1].full_name, container, axis=1)
        else:
            apply_identity(scope, scores,
                           operator.outputs[1].full_name, container)
        proba = operator.outputs[1].full_name

    container.add_node('ArgMax', proba,
                       predicted_label_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, predicted_label_name],
        final_label_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    if class_type == onnx_proto.TensorProto.INT32:
        reshaped_final_label_name = scope.get_unique_variable_name(
            'reshaped_final_label')

        apply_reshape(scope, final_label_name, reshaped_final_label_name,
                      container, desired_shape=(-1,))
        apply_cast(scope, reshaped_final_label_name,
                   operator.outputs[0].full_name, container,
                   to=onnx_proto.TensorProto.INT64)
    else:
        apply_reshape(scope, final_label_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=(-1,))


register_converter('SklearnSGDClassifier',
                   convert_sklearn_sgd_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator


def calculate_sklearn_zipmap(operator):
    if (len(operator.inputs) != len(operator.outputs) or
            len(operator.inputs) not in (1, 2)):
        raise RuntimeError(
            "SklearnZipMap expects the same number of inputs and outputs.")
    if len(operator.inputs) == 2:
        operator.outputs[0].type = operator.inputs[0].type.__class__(
            operator.inputs[0].type.shape)
        if operator.outputs[1].type is not None:
            operator.outputs[1].type.element_type.value_type = \
                operator.inputs[1].type.__class__([])


def calculate_sklearn_zipmap_columns(operator):
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = operator.inputs[0].type.__class__(
        operator.inputs[0].type.shape)
    for i in range(1, len(operator.outputs)):
        operator.outputs[i].type.shape = [N]


register_shape_calculator('SklearnZipMap', calculate_sklearn_zipmap)
register_shape_calculator(
    'SklearnZipMapColumns', calculate_sklearn_zipmap_columns)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._apply_operation import (
    apply_cast, apply_add, apply_sqrt, apply_div, apply_sub,
    apply_reshape)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, DoubleTensorType,
    guess_numpy_type, guess_proto_type)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto
from ..algebra.onnx_ops import (
    OnnxAdd, OnnxCast, OnnxExp, OnnxIdentity, OnnxMatMul,
    OnnxReshape, OnnxSigmoid)


def convert_sklearn_linear_regressor(scope: Scope, operator: Operator,
                                     container: ModelComponentContainer):
    op = operator.raw_operator
    use_linear_op = container.is_allowed({'LinearRegressor'})

    if (not use_linear_op or
            type(operator.inputs[0].type) in (DoubleTensorType, )):
        proto_dtype = guess_proto_type(operator.inputs[0].type)
        coef = scope.get_unique_variable_name('coef')
        if len(op.coef_.shape) == 1:
            model_coef = op.coef_.reshape((-1, 1))
        else:
            model_coef = op.coef_.T
        container.add_initializer(
            coef, proto_dtype, model_coef.shape, model_coef.ravel().tolist())
        intercept = scope.get_unique_variable_name('intercept')
        value_intercept = op.intercept_.reshape((-1, ))
        container.add_initializer(
            intercept, proto_dtype, value_intercept.shape,
            value_intercept.ravel().tolist())
        multiplied = scope.get_unique_variable_name('multiplied')
        container.add_node(
            'MatMul', [operator.inputs[0].full_name, coef], multiplied,
            name=scope.get_unique_operator_name('MatMul'))
        resh = scope.get_unique_variable_name('resh')
        apply_add(scope, [multiplied, intercept],
                  resh, container)
        last_dim = 1 if len(model_coef.shape) == 1 else model_coef.shape[-1]
        apply_reshape(scope, resh, operator.outputs[0].full_name,
                      container, desired_shape=(-1, last_dim))
        return

    op_type = 'LinearRegressor'
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype not in (np.float32, np.float64):
        dtype = np.float32
    attrs = {'name': scope.get_unique_operator_name(op_type)}
    attrs['coefficients'] = op.coef_.astype(dtype).ravel()
    attrs['intercepts'] = np.array([op.intercept_], dtype=dtype).ravel()
    if len(op.coef_.shape) == 2:
        attrs['targets'] = op.coef_.shape[0]

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container,
                   to=(onnx_proto.TensorProto.DOUBLE
                       if dtype == np.float64
                       else onnx_proto.TensorProto.FLOAT))
        input_name = cast_input_name
    container.add_node(op_type, input_name,
                       operator.outputs[0].full_name, op_domain='ai.onnx.ml',
                       **attrs)


def convert_sklearn_bayesian_ridge(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    convert_sklearn_linear_regressor(scope, operator, container)

    op = operator.raw_operator
    options = container.get_options(op, dict(return_std=False))
    return_std = options['return_std']
    if not return_std:
        return

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if op.normalize:
        # if self.normalize:
        #     X = (X - self.X_offset_) / self.X_scale_
        offset = scope.get_unique_variable_name('offset')
        container.add_initializer(
            offset, proto_dtype, op.X_offset_.shape,
            op.X_offset_.ravel().tolist())
        scale = scope.get_unique_variable_name('scale')
        container.add_initializer(
            scale, proto_dtype, op.X_scale_.shape,
            op.X_scale_.ravel().tolist())
        centered = scope.get_unique_variable_name('centered')
        apply_sub(scope, [operator.inputs[0].full_name, offset],
                  centered, container)
        scaled = scope.get_unique_variable_name('scaled')
        apply_div(scope, [centered, scale], scaled, container)
        input_name = scaled
    else:
        input_name = operator.inputs[0].full_name

    # sigmas_squared_data = (np.dot(X, self.sigma_) * X).sum(axis=1)
    sigma = scope.get_unique_variable_name('sigma')
    container.add_initializer(
        sigma, proto_dtype, op.sigma_.shape, op.sigma_.ravel().tolist())
    sigmaed0 = scope.get_unique_variable_name('sigma0')
    container.add_node(
        'MatMul', [input_name, sigma], sigmaed0,
        name=scope.get_unique_operator_name('MatMul'))
    sigmaed = scope.get_unique_variable_name('sigma')
    if container.target_opset < 13:
        container.add_node(
            'ReduceSum', sigmaed0, sigmaed, axes=[1],
            name=scope.get_unique_operator_name('ReduceSum'))
    else:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            'ReduceSum', [sigmaed0, axis_name], sigmaed,
            name=scope.get_unique_operator_name('ReduceSum'))

    # y_std = np.sqrt(sigmas_squared_data + (1. / self.alpha_))
    # return y_mean, y_std
    std0 = scope.get_unique_variable_name('std0')
    alphainv = scope.get_unique_variable_name('alphainv')
    container.add_initializer(alphainv, proto_dtype, [1], [1 / op.alpha_])
    apply_add(scope, [sigmaed, alphainv], std0, container)
    apply_sqrt(scope, std0, operator.outputs[1].full_name, container)


def convert_sklearn_poisson_regressor(scope: Scope, operator: Operator,
                                      container: ModelComponentContainer):
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    if type(X.type) == Int64TensorType:
        input_var = OnnxCast(X, to=np.float32, op_version=opv)
    else:
        input_var = X

    intercept = (op.intercept_.astype(dtype) if len(op.intercept_.shape) > 0
                 else np.array([op.intercept_], dtype=dtype))
    eta = OnnxAdd(
        OnnxMatMul(input_var, op.coef_.astype(dtype), op_version=opv),
        intercept, op_version=opv)

    if hasattr(op, "_link_instance"):
        # scikit-learn < 1.1
        from sklearn.linear_model._glm.link import (
            IdentityLink, LogLink, LogitLink)
        if isinstance(op._link_instance, IdentityLink):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(op._link_instance, LogLink):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(op._link_instance, LogitLink):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                "Unexpected type %r for _link_instance "
                "in operator type %r." % (
                    type(op._link_instance), type(op)))
    else:
        # scikit-learn >= 1.1
        from sklearn._loss.loss import (
            AbsoluteError,
            HalfBinomialLoss,
            HalfGammaLoss,
            HalfPoissonLoss,
            HalfSquaredError,
            HalfTweedieLoss,
            HalfTweedieLossIdentity,
            PinballLoss
        )
        loss = op._get_loss()
        if isinstance(
            loss,
            (AbsoluteError, HalfSquaredError,
             HalfTweedieLossIdentity, PinballLoss)):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(loss, (HalfPoissonLoss, HalfGammaLoss,
                               HalfTweedieLoss)):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(loss, HalfBinomialLoss):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                f"Unexpected type of link for {loss!r} loss "
                "in operator type {op!r}.")

    last_dim = 1 if len(op.coef_.shape) == 1 else op.coef_.shape[-1]
    final = OnnxReshape(Y, np.array([-1, last_dim], dtype=np.int64),
                        op_version=opv, output_names=out[:1])
    final.add_to(scope, container)


register_converter('SklearnLinearRegressor', convert_sklearn_linear_regressor)
register_converter('SklearnLinearSVR', convert_sklearn_linear_regressor)
register_converter('SklearnBayesianRidge', convert_sklearn_bayesian_ridge,
                   options={'return_std': [True, False]})
register_converter('SklearnPoissonRegressor',
                   convert_sklearn_poisson_regressor)
register_converter('SklearnTweedieRegressor',
                   convert_sklearn_poisson_regressor)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from onnx.helper import make_tensor
from ..algebra.onnx_ops import (
    OnnxAdd,
    OnnxArgMax,
    OnnxArrayFeatureExtractor,
    OnnxCast,
    OnnxConcat,
    OnnxDiv,
    OnnxEqual,
    OnnxFlatten,
    OnnxIdentity,
    OnnxLess,
    OnnxMatMul,
    OnnxMax,
    OnnxMul,
    OnnxNeg,
    OnnxNot,
    OnnxReciprocal,
    OnnxReduceMean,
    OnnxReduceSumApi11,
    OnnxReshapeApi13,
    OnnxShape,
    OnnxSqueezeApi11,
    OnnxSub,
    OnnxTopK_1,
    OnnxTranspose,
)
try:
    from ..algebra.onnx_ops import (
        OnnxConstantOfShape,
        OnnxCumSum,
        OnnxIsNaN,
        OnnxWhere,
    )
except ImportError:
    OnnxConstantOfShape = None
    OnnxCumSum = None
    OnnxIsNaN = None
    OnnxWhere = None
try:
    from ..algebra.onnx_ops import OnnxTopK_10
except ImportError:
    OnnxTopK_10 = None
try:
    from ..algebra.onnx_ops import OnnxTopK_11
except ImportError:
    OnnxTopK_11 = None
from ..algebra.complex_functions import onnx_cdist, _onnx_cdist_sqeuclidean
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    Int64TensorType, DoubleTensorType,
    guess_numpy_type, guess_proto_type)
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto
from ._gp_kernels import py_make_float_array


def onnx_nearest_neighbors_indices_k(X, Y, k, metric='euclidean', dtype=None,
                                     op_version=None, keep_distances=False,
                                     optim=None, **kwargs):
    """
    Retrieves the nearest neigbours *ONNX*.
    :param X: features or *OnnxOperatorMixin*
    :param Y: neighbours or *OnnxOperatorMixin*
    :param k: number of neighbours to retrieve
    :param metric: requires metric
    :param dtype: numerical type
    :param op_version: opset version
    :param keep_distance: returns the distances as well (second position)
    :param optim: implements specific optimisations,
        ``'cdist'`` replaces *Scan* operator by operator *CDist*
    :param kwargs: additional parameters for function @see fn onnx_cdist
    :return: top indices, top distances
    """
    kwargs_dist = {k: v for k, v in kwargs.items() if k == 'p'}
    kwargs_topk = {k: v for k, v in kwargs.items() if k != 'p'}
    if optim == 'cdist':
        from skl2onnx.algebra.custom_ops import OnnxCDist
        dist = OnnxCDist(X, Y, metric=metric, op_version=op_version,
                         **kwargs_dist)
    elif optim is None:
        dim_in = Y.shape[1] if hasattr(Y, 'shape') else None
        dim_out = Y.shape[0] if hasattr(Y, 'shape') else None
        dist = onnx_cdist(X, Y, metric=metric, dtype=dtype,
                          op_version=op_version,
                          dim_in=dim_in, dim_out=dim_out,
                          **kwargs_dist)
    else:
        raise ValueError("Unknown optimisation '{}'.".format(optim))
    if op_version < 10:
        neg_dist = OnnxMul(dist, np.array([-1], dtype=dtype),
                           op_version=op_version)
        node = OnnxTopK_1(neg_dist, k=k, op_version=1, **kwargs_topk)
    elif op_version < 11:
        neg_dist = OnnxMul(dist, np.array([-1], dtype=dtype),
                           op_version=op_version)
        node = OnnxTopK_10(neg_dist, np.array([k], dtype=np.int64),
                           op_version=10, **kwargs_topk)
    else:
        node = OnnxTopK_11(dist, np.array([k], dtype=np.int64),
                           largest=0, sorted=1,
                           op_version=11, **kwargs_topk)
        if keep_distances:
            return (node[1], OnnxMul(
                node[0], np.array([-1], dtype=dtype), op_version=op_version))
    if keep_distances:
        return (node[1], node[0])
    return node[1]


def onnx_nearest_neighbors_indices_radius(
        X, Y, radius, metric='euclidean', dtype=None, op_version=None,
        keep_distances=False, optim=None, proto_dtype=None, **kwargs):
    """
    Retrieves the nearest neigbours *ONNX*.
    :param X: features or *OnnxOperatorMixin*
    :param Y: neighbours or *OnnxOperatorMixin*
    :param radius: radius
    :param metric: requires metric
    :param dtype: numerical type
    :param op_version: opset version
    :param keep_distance: returns the distances as well (second position)
    :param optim: implements specific optimisations,
        ``'cdist'`` replaces *Scan* operator by operator *CDist*
    :param kwargs: additional parameters for function @see fn onnx_cdist
    :return: 3 squares matrices, indices or -1, distance or 0,
        based on the fact that the distance is below the radius,
        binary weights
    """
    opv = op_version
    if optim == 'cdist':
        from skl2onnx.algebra.custom_ops import OnnxCDist
        dist = OnnxCDist(X, Y, metric=metric, op_version=op_version,
                         **kwargs)
    elif optim is None:
        dim_in = Y.shape[1] if hasattr(Y, 'shape') else None
        dim_out = Y.shape[0] if hasattr(Y, 'shape') else None
        dist = onnx_cdist(X, Y, metric=metric, dtype=dtype,
                          op_version=op_version,
                          dim_in=dim_in, dim_out=dim_out,
                          **kwargs)
    else:
        raise ValueError("Unknown optimisation '{}'.".format(optim))

    less = OnnxLess(dist, np.array([radius], dtype=dtype), op_version=opv)
    less.set_onnx_name_prefix('cond')
    shape = OnnxShape(dist, op_version=opv)
    zero = OnnxCast(
        OnnxConstantOfShape(shape, op_version=opv),
        op_version=opv, to=proto_dtype)
    tensor_value = py_make_float_array(-1, dtype=np.float32, as_tensor=True)
    minus = OnnxCast(
        OnnxConstantOfShape(
            shape, op_version=opv, value=tensor_value),
        op_version=opv, to=onnx_proto.TensorProto.INT64)
    minus_range = OnnxAdd(
        OnnxNeg(
            OnnxCumSum(minus, np.array([1], dtype=np.int64), op_version=opv),
            op_version=opv),
        minus, op_version=opv)
    minus_range.set_onnx_name_prefix('arange')

    dist_only = OnnxWhere(less, dist, zero, op_version=opv)
    dist_only.set_onnx_name_prefix('nndist')
    indices = OnnxWhere(less, minus_range, minus, op_version=opv)
    indices.set_onnx_name_prefix('nnind')
    binary = OnnxCast(less, to=proto_dtype, op_version=opv)
    binary.set_onnx_name_prefix('nnbin')
    return indices, dist_only, binary


def _convert_nearest_neighbors(operator, container, k=None, radius=None):
    """
    Common parts to regressor and classifier. Let's denote
    *N* as the number of observations, *k*
    the number of neighbours. It returns
    the following intermediate results:

    top_indices: [N, k] (int64), best indices for
        every observation
    top_distances: [N, k] (dtype), float distances
        for every observation, it can be None
        if the weights are uniform
    top_labels: [N, k] (label type), labels
        associated to every top index
    weights: [N, k] (dtype), if top_distances is not None,
        returns weights
    norm: [N] (dtype), if top_distances is not None,
        returns normalized weights
    axis: 1 if there is one dimension only, 2 if
        this is a multi-regression or a multi classification
    """
    X = operator.inputs[0]
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    proto_type = guess_proto_type(X.type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT

    if isinstance(X.type, Int64TensorType):
        X = OnnxCast(X, to=proto_type, op_version=opv)

    options = container.get_options(op, dict(optim=None))

    single_reg = (not hasattr(op, '_y') or len(op._y.shape) == 1 or
                  len(op._y.shape) == 2 and op._y.shape[1] == 1)
    ndim = 1 if single_reg else op._y.shape[1]

    metric = (op.effective_metric_ if hasattr(op, 'effective_metric_') else
              op.metric)
    neighb = op._fit_X.astype(dtype)

    if (hasattr(op, 'n_neighbors') and op.n_neighbors is not None and
            hasattr(op, 'radius') and op.radius is not None):
        raise RuntimeError(
            "The model defines radius and n_neighbors at the "
            "same time ({} and {}). "
            "This case is not supported.".format(
                op.radius, op.n_neighbors))

    if hasattr(op, 'n_neighbors') and op.n_neighbors is not None:
        k = op.n_neighbors if k is None else k
        radius = None
    elif hasattr(op, 'radius') and op.radius is not None:
        k = None
        radius = op.radius if radius is None else radius
    else:
        raise RuntimeError(
            "Cannot convert class '{}'.".format(op.__class__.__name__))

    training_labels = op._y if hasattr(op, '_y') else None
    distance_kwargs = {}
    if metric == 'minkowski':
        if op.p != 2:
            distance_kwargs['p'] = op.p
        else:
            metric = "euclidean"

    weights = op.weights if hasattr(op, 'weights') else 'distance'
    binary = None
    if weights == 'uniform' and radius is None:
        top_indices = onnx_nearest_neighbors_indices_k(
            X, neighb, k, metric=metric, dtype=dtype,
            op_version=opv, optim=options.get('optim', None),
            **distance_kwargs)
        top_distances = None
    elif radius is not None:
        three = onnx_nearest_neighbors_indices_radius(
            X, neighb, radius, metric=metric, dtype=dtype,
            op_version=opv, keep_distances=True,
            proto_dtype=proto_type,
            optim=options.get('optim', None),
            **distance_kwargs)
        top_indices, top_distances, binary = three
    elif weights == 'distance':
        top_indices, top_distances = onnx_nearest_neighbors_indices_k(
            X, neighb, k, metric=metric, dtype=dtype,
            op_version=opv, keep_distances=True,
            optim=options.get('optim', None),
            **distance_kwargs)
    else:
        raise RuntimeError(
            "Unable to convert KNeighborsRegressor when weights is callable.")

    if training_labels is not None:
        if ndim > 1:
            training_labels = training_labels.T
            axis = 2
        else:
            training_labels = training_labels.ravel()
            axis = 1
        if opv >= 9:
            kor = k if k is not None else training_labels.shape[-1]
            if ndim > 1:
                shape = np.array([ndim, -1, kor], dtype=np.int64)
            else:
                shape = np.array([-1, kor], dtype=np.int64)
        else:
            raise RuntimeError(
                "Conversion of a KNeighborsRegressor for multi regression "
                "requires opset >= 9.")

        if training_labels.dtype == np.int32:
            training_labels = training_labels.astype(np.int64)
        flattened = OnnxFlatten(top_indices, op_version=opv)
        extracted = OnnxArrayFeatureExtractor(
            training_labels, flattened, op_version=opv)
        reshaped = OnnxReshapeApi13(extracted, shape, op_version=opv)

        if ndim > 1:
            reshaped = OnnxTranspose(reshaped, op_version=opv, perm=[1, 0, 2])
        reshaped.set_onnx_name_prefix('knny')

    else:
        reshaped = None
        axis = 1

    if binary is not None:
        if op.weights == 'uniform':
            wei = binary
        else:
            modified = OnnxMax(top_distances, np.array([1e-6], dtype=dtype),
                               op_version=opv)
            wei = OnnxMul(binary, OnnxReciprocal(modified, op_version=opv),
                          op_version=opv)
        norm = OnnxReduceSumApi11(wei, op_version=opv, axes=[1], keepdims=0)
    elif top_distances is not None:
        modified = OnnxMax(top_distances, np.array([1e-6], dtype=dtype),
                           op_version=opv)
        wei = OnnxReciprocal(modified, op_version=opv)
        norm = OnnxReduceSumApi11(wei, op_version=opv, axes=[1], keepdims=0)
    else:
        norm = None
        wei = None

    if wei is not None:
        wei.set_onnx_name_prefix('wei')
    if norm is not None:
        norm.set_onnx_name_prefix('norm')
    return top_indices, top_distances, reshaped, wei, norm, axis


def convert_nearest_neighbors_regressor(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    """
    Converts *KNeighborsRegressor* into *ONNX*.
    The converted model may return different predictions depending
    on how the runtime select the topk element.
    *scikit-learn* uses function `argpartition
    <https://docs.scipy.org/doc/numpy/reference/
    generated/numpy.argpartition.html>`_ which keeps the
    original order of the elements.
    """
    many = _convert_nearest_neighbors(operator, container)
    _, top_distances, reshaped, wei, norm, axis = many
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT

    opv = container.target_opset
    out = operator.outputs

    reshaped_cast = OnnxCast(
        reshaped, to=proto_type, op_version=opv)
    if top_distances is not None:
        # Multi-target
        if (hasattr(operator.raw_operator, '_y') and
                len(operator.raw_operator._y.shape) > 1 and
                operator.raw_operator._y.shape[1] > 1):
            rs = OnnxTranspose(reshaped_cast, perm=[1, 0, 2],
                               op_version=opv)
            weighted_rs = OnnxMul(rs, wei, op_version=opv)
            weighted = OnnxTranspose(weighted_rs, perm=[1, 0, 2],
                                     op_version=opv)

            if OnnxIsNaN is not None:
                # This steps sometimes produces nan (bug in onnxuntime)
                # They are replaced by null values.
                isnan = OnnxIsNaN(weighted, op_version=opv)
                shape = OnnxShape(weighted, op_version=opv)
                csts0 = OnnxConstantOfShape(shape, op_version=opv)
                weighted = OnnxWhere(isnan, csts0, weighted, op_version=opv)
                # Back to original plan.

            res = OnnxReduceSumApi11(weighted, axes=[axis], op_version=opv,
                                     keepdims=0)
            norm2 = OnnxReshapeApi13(norm, np.array([-1, 1], dtype=np.int64),
                                     op_version=opv)
            res = OnnxDiv(res, norm2, op_version=opv, output_names=out)
        else:
            weighted = OnnxMul(reshaped_cast, wei, op_version=opv)
            res = OnnxReduceSumApi11(weighted, axes=[axis], op_version=opv,
                                     keepdims=0)
            res.set_onnx_name_prefix('final')
            if opv >= 12:
                shape = OnnxShape(res, op_version=opv)
                norm = OnnxReshapeApi13(norm, shape, op_version=opv)
                norm.set_onnx_name_prefix('normr')
            res = OnnxDiv(res, norm, op_version=opv)
            res = OnnxReshapeApi13(res, np.array([-1, 1], dtype=np.int64),
                                   output_names=out, op_version=opv)
    else:
        if (hasattr(operator.raw_operator, '_y') and
                len(np.squeeze(operator.raw_operator._y).shape) == 1):
            keepdims = 1
        elif operator.raw_operator.n_neighbors == 1:
            keepdims = 0
        else:
            keepdims = 0
        res = OnnxReduceMean(reshaped_cast, axes=[axis], op_version=opv,
                             keepdims=keepdims, output_names=out)
    res.add_to(scope, container)


def get_proba_and_label(container, nb_classes, reshaped,
                        wei, axis, opv, proto_type, keep_axis=True):
    """
    This function calculates the label by choosing majority label
    amongst the nearest neighbours.
    """
    conc = []
    for cl in range(nb_classes):
        cst = np.array([cl], dtype=np.int64)
        mat_cast = OnnxCast(
            OnnxEqual(reshaped, cst, op_version=opv),
            op_version=opv, to=proto_type)
        if wei is not None:
            if not keep_axis:
                mat_cast = OnnxSqueezeApi11(mat_cast, axes=[-1],
                                            op_version=opv)
            mat_cast = OnnxMul(mat_cast, wei, op_version=opv)
        wh = OnnxReduceSumApi11(mat_cast, axes=[1], op_version=opv)
        conc.append(wh)
    all_together = OnnxConcat(*conc, axis=1, op_version=opv)
    sum_prob = OnnxReduceSumApi11(
        all_together, axes=[1], op_version=opv, keepdims=1)
    res = OnnxArgMax(all_together, axis=axis, op_version=opv,
                     keepdims=0)
    return all_together, sum_prob, res


def convert_nearest_neighbors_classifier(scope: Scope, operator: Operator,
                                         container: ModelComponentContainer):
    """
    Converts *KNeighborsClassifier* into *ONNX*.
    The converted model may return different predictions depending
    on how the runtime select the topk element.
    *scikit-learn* uses function `argpartition
    <https://docs.scipy.org/doc/numpy/reference/
    generated/numpy.argpartition.html>`_ which keeps the
    original order of the elements.
    """
    many = _convert_nearest_neighbors(operator, container)
    _, __, reshaped, wei, ___, axis = many

    opv = container.target_opset
    out = operator.outputs
    op = operator.raw_operator
    nb_classes = len(op.classes_)
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT

    if axis == 0:
        raise RuntimeError(
            "Binary classification not implemented in scikit-learn. "
            "Check this code is not reused for other libraries.")
    classes = get_label_classes(scope, op)
    if hasattr(classes, 'dtype') and np.issubdtype(classes.dtype, np.floating):
        classes = classes.astype(np.int32)
        is_integer = True
    elif isinstance(classes[0], (int, np.int32, np.int64)):
        is_integer = True
    else:
        is_integer = False
    if (isinstance(op.classes_, list)
            and isinstance(op.classes_[0], np.ndarray)):
        # Multi-label
        out_labels, out_probas = [], []
        for index, cur_class in enumerate(op.classes_):
            transpose_result = OnnxTranspose(
                reshaped, op_version=opv, perm=[0, 2, 1])
            extracted_name = OnnxArrayFeatureExtractor(
                transpose_result, np.array([index], dtype=np.int64),
                op_version=opv)
            extracted_name.set_onnx_name_prefix('tr%d' % index)
            all_together, sum_prob, res = get_proba_and_label(
                container, len(cur_class), extracted_name,
                wei, 1, opv, proto_type, keep_axis=False)
            probas = OnnxDiv(all_together, sum_prob, op_version=opv)
            res_name = OnnxArrayFeatureExtractor(
                cur_class, res, op_version=opv)
            res_name.set_onnx_name_prefix('div%d' % index)
            reshaped_labels = OnnxReshapeApi13(
                res_name, np.array([-1, 1], dtype=np.int64), op_version=opv)
            reshaped_probas = OnnxReshapeApi13(
                probas, np.array([1, -1, len(cur_class)], dtype=np.int64),
                op_version=opv)
            out_labels.append(reshaped_labels)
            out_probas.append(reshaped_probas)
        concatenated_labels = OnnxConcat(
            *out_labels, axis=1, op_version=opv)
        final_proba = OnnxConcat(
            *out_probas, axis=0, output_names=out[1:], op_version=opv)
        final_label = OnnxCast(
            concatenated_labels, to=onnx_proto.TensorProto.INT64,
            output_names=out[:1], op_version=opv)
        final_label.add_to(scope, container)
        final_proba.add_to(scope, container)
    else:
        all_together, sum_prob, res = get_proba_and_label(
            container, nb_classes, reshaped, wei, axis, opv, proto_type)
        probas = OnnxDiv(all_together, sum_prob, op_version=opv,
                         output_names=out[1:])
        probas.set_onnx_name_prefix('bprob')
        res_name = OnnxArrayFeatureExtractor(classes, res, op_version=opv)
        if is_integer:
            res_name = OnnxCast(
                res_name, to=onnx_proto.TensorProto.INT64, op_version=opv)
        out_labels = OnnxReshapeApi13(res_name, np.array([-1], dtype=np.int64),
                                      output_names=out[:1], op_version=opv)
        out_labels.set_onnx_name_prefix('blab')
        out_labels.add_to(scope, container)
        probas.add_to(scope, container)


def convert_nearest_neighbors_transform(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    """
    Converts *NearestNeighbors* into *ONNX*.
    """
    many = _convert_nearest_neighbors(operator, container)
    top_indices, top_distances = many[:2]
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    out = operator.outputs

    ind = OnnxIdentity(top_indices, output_names=out[:1],
                       op_version=container.target_opset)
    dist = OnnxMul(
        top_distances, np.array([-1], dtype=dtype),
        output_names=out[1:], op_version=container.target_opset)

    dist.add_to(scope, container)
    ind.add_to(scope, container)


def convert_k_neighbours_transformer(scope: Scope, operator: Operator,
                                     container: ModelComponentContainer):
    """
    Converts *KNeighborsTransformer* into *ONNX*.
    """
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    transformer_op = operator.raw_operator
    op_version = container.target_opset
    k = (transformer_op.n_neighbors + 1 if transformer_op.mode == 'distance'
         else transformer_op.n_neighbors)
    out = operator.outputs

    many = _convert_nearest_neighbors(
        operator, container, k=k)
    top_indices, top_dist = many[:2]
    top_dist = (
        OnnxReshapeApi13(
            OnnxMul(top_dist, np.array([-1], dtype=dtype),
                    op_version=op_version),
            np.array([-1, 1, k], dtype=np.int64),
            op_version=op_version)
        if transformer_op.mode == 'distance' else None)
    fit_samples_indices = np.array(
        np.arange(transformer_op.n_samples_fit_).reshape((1, -1, 1)),
        dtype=np.int64)
    reshaped_ind = OnnxReshapeApi13(
        top_indices, np.array([-1, 1, k], dtype=np.int64),
        op_version=op_version)
    comparison_res = OnnxCast(
        OnnxEqual(fit_samples_indices, reshaped_ind, op_version=op_version),
        op_version=op_version, to=proto_type)
    if top_dist:
        comparison_res = OnnxMul(
            comparison_res, top_dist, op_version=op_version)
    res = OnnxReduceSumApi11(
        comparison_res, op_version=op_version, axes=[2],
        keepdims=0, output_names=out[:1])
    res.add_to(scope, container)


def _nan_euclidean_distance(
        container, model, input_name, op_version, optim, dtype, proto_type):
    training_data = model._fit_X.astype(dtype)
    shape = OnnxShape(input_name, op_version=op_version)
    zero = OnnxConstantOfShape(
        shape, value=make_tensor("value", proto_type, (1, ), [0]),
        op_version=op_version)
    missing_input_name = OnnxIsNaN(input_name, op_version=op_version)
    masked_input_name = OnnxWhere(missing_input_name, zero, input_name,
                                  op_version=op_version)
    missing_y = np.isnan(training_data)
    training_data[missing_y] = 0
    d_in = training_data.shape[1] if hasattr(training_data, 'shape') else None
    d_out = training_data.shape[0] if hasattr(training_data, 'shape') else None

    if optim is None:
        dist = _onnx_cdist_sqeuclidean(
            masked_input_name, training_data, dtype=dtype,
            op_version=container.target_opset, dim_in=d_in, dim_out=d_out)
    elif optim == 'cdist':
        from skl2onnx.algebra.custom_ops import OnnxCDist
        dist = OnnxCDist(
            masked_input_name, training_data, metric='sqeuclidean',
            op_version=container.target_opset)
    else:
        raise RuntimeError("Unexpected optimization '{}'.".format(optim))
    dist1 = OnnxMatMul(
        OnnxMul(masked_input_name, masked_input_name, op_version=op_version),
        missing_y.T.astype(dtype), op_version=op_version)
    dist2 = OnnxMatMul(
        OnnxCast(missing_input_name, to=proto_type,
                 op_version=op_version),
        (training_data * training_data).T.astype(dtype),
        op_version=op_version)
    distances = OnnxSub(dist, OnnxAdd(dist1, dist2, op_version=op_version),
                        op_version=op_version)
    present_x = OnnxSub(
        np.array([1], dtype=dtype),
        OnnxCast(missing_input_name, to=proto_type,
                 op_version=op_version),
        op_version=op_version)
    present_y = (1. - missing_y).astype(dtype)
    present_count = OnnxMatMul(
        present_x, present_y.T.astype(dtype), op_version=op_version)
    present_count = OnnxMax(np.array([1], dtype=dtype),
                            present_count, op_version=op_version)
    dist = OnnxDiv(distances, present_count, op_version=op_version)
    return OnnxMul(
        dist, np.array([d_in], dtype=dtype),
        op_version=op_version), missing_input_name


def _nearest_neighbours(container, model, input_name,
                        op_version, optim, dtype, proto_type, **kwargs):
    dist, missing_input_name = _nan_euclidean_distance(
        container, model, input_name, op_version, optim, dtype,
        proto_type)
    if op_version < 10:
        neg_dist = OnnxMul(dist, np.array(
            [-1], dtype=dtype), op_version=op_version)
        node = OnnxTopK_1(
            neg_dist, k=model.n_neighbors, op_version=1, **kwargs)
    elif op_version < 11:
        neg_dist = OnnxMul(dist, np.array(
            [-1], dtype=dtype), op_version=op_version)
        node = OnnxTopK_10(
            neg_dist, np.array([model.n_neighbors], dtype=np.int64),
            op_version=10, **kwargs)
    else:
        node = OnnxTopK_11(
            dist, np.array([model.n_neighbors], dtype=np.int64),
            largest=0, sorted=1, op_version=11, **kwargs)
    return node[1], missing_input_name


def convert_knn_imputer(scope: Scope, operator: Operator,
                        container: ModelComponentContainer):
    """
    Converts *KNNImputer* into *ONNX*.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    proto_type = guess_proto_type(operator.inputs[0].type)
    if proto_type != onnx_proto.TensorProto.DOUBLE:
        proto_type = onnx_proto.TensorProto.FLOAT
    knn_op = operator.raw_operator
    if knn_op.metric != 'nan_euclidean':
        raise RuntimeError(
            "Unable to convert KNNImputer when metric is callable.")
    if knn_op.weights not in ('uniform', 'distance'):
        raise RuntimeError(
            "Unable to convert KNNImputer when weights is callable.")
    if knn_op.weights == 'distance':
        raise NotImplementedError(
            'KNNImputer with distance as metric is not supported, '
            'you may raise an issue at '
            'https://github.com/onnx/sklearn-onnx/issues.')
    options = container.get_options(knn_op, dict(optim=None))
    op_version = container.target_opset
    input_name = operator.inputs[0]
    training_data = knn_op._fit_X.astype(dtype)
    training_data[np.isnan(training_data)] = 0
    out = operator.outputs
    top_indices, missing_input_name = _nearest_neighbours(
        container, knn_op, input_name, op_version, options['optim'],
        dtype, proto_type)
    flattened = OnnxFlatten(top_indices, op_version=op_version)
    extracted = OnnxArrayFeatureExtractor(
        training_data.T, flattened, op_version=op_version)
    reshaped = OnnxReshapeApi13(
        extracted, np.array([training_data.shape[1], -1, knn_op.n_neighbors],
                            dtype=np.int64),
        op_version=op_version)
    transpose_result = OnnxTranspose(
        reshaped, op_version=op_version, perm=[1, 2, 0])
    reduced = OnnxReduceSumApi11(
        transpose_result, op_version=op_version, axes=[1], keepdims=0)
    cast_res = OnnxCast(
        OnnxCast(transpose_result, to=onnx_proto.TensorProto.BOOL,
                 op_version=op_version),
        to=proto_type, op_version=op_version)
    deno = OnnxReduceSumApi11(
        cast_res, op_version=op_version, axes=[1], keepdims=0)
    deno_updated = OnnxAdd(
        deno, OnnxCast(
            OnnxNot(OnnxCast(deno, to=onnx_proto.TensorProto.BOOL,
                             op_version=op_version), op_version=op_version),
            to=proto_type, op_version=op_version),
        op_version=op_version)
    imputed_out = OnnxWhere(
        missing_input_name,
        OnnxDiv(reduced, deno_updated, op_version=op_version), input_name,
        output_names=out[:1], op_version=op_version)
    imputed_out.add_to(scope, container)


def convert_nca(scope: Scope, operator: Operator,
                container: ModelComponentContainer):
    """
    Converts *NeighborhoodComponentsAnalysis* into *ONNX*.
    """
    X = operator.inputs[0]
    nca_op = operator.raw_operator
    op_version = container.target_opset
    out = operator.outputs
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    components = nca_op.components_.T.astype(dtype)

    if isinstance(X.type, Int64TensorType):
        X = OnnxCast(X, to=onnx_proto.TensorProto.FLOAT, op_version=op_version)
    elif isinstance(X.type, DoubleTensorType):
        components = OnnxCast(
            components, to=onnx_proto.TensorProto.DOUBLE,
            op_version=op_version)
    else:
        components = components.astype(dtype)
    res = OnnxMatMul(
        X, components,
        output_names=out[:1], op_version=op_version)
    res.add_to(scope, container)


register_converter(
    'SklearnKNeighborsClassifier', convert_nearest_neighbors_classifier,
    options={'zipmap': [True, False, 'columns'],
             'nocl': [True, False],
             'raw_scores': [True, False],
             'output_class_labels': [False, True],
             'optim': [None, 'cdist']})
register_converter(
    'SklearnRadiusNeighborsClassifier', convert_nearest_neighbors_classifier,
    options={'zipmap': [True, False, 'columns'],
             'nocl': [True, False],
             'raw_scores': [True, False],
             'output_class_labels': [False, True],
             'optim': [None, 'cdist']})
register_converter(
    'SklearnKNeighborsRegressor', convert_nearest_neighbors_regressor,
    options={'optim': [None, 'cdist']})
register_converter(
    'SklearnRadiusNeighborsRegressor', convert_nearest_neighbors_regressor,
    options={'optim': [None, 'cdist']})
register_converter(
    'SklearnKNeighborsTransformer', convert_k_neighbours_transformer,
    options={'optim': [None, 'cdist']})
register_converter(
    'SklearnNearestNeighbors', convert_nearest_neighbors_transform,
    options={'optim': [None, 'cdist']})
register_converter(
    'SklearnKNNImputer', convert_knn_imputer,
    options={'optim': [None, 'cdist']})
register_converter(
    'SklearnNeighborhoodComponentsAnalysis', convert_nca)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_text_vectorizer_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. Map ---> [1, C]

    C is the total number of allowed keys in the input dictionary.
    '''
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)

    C = max(operator.raw_operator.vocabulary_.values()) + 1
    operator.outputs[0].type.shape = [None, C]


register_shape_calculator('SklearnCountVectorizer',
                          calculate_sklearn_text_vectorizer_output_shapes)
register_shape_calculator('SklearnTfidfVectorizer',
                          calculate_sklearn_text_vectorizer_output_shapes)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator


def random_projection_shape_calculator(operator):
    """Shape calculator for PowerTransformer"""
    inputs = operator.inputs[0]
    op = operator.raw_operator
    n = inputs.get_first_dimension()
    c = op.components_.shape[0]
    operator.outputs[0].type.shape = [n, c]


register_shape_calculator('SklearnGaussianRandomProjection',
                          random_projection_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_label_encoder(scope: Scope, operator: Operator,
                                  container: ModelComponentContainer):
    op = operator.raw_operator
    op_type = 'LabelEncoder'
    attrs = {'name': scope.get_unique_operator_name(op_type)}
    classes = op.classes_
    if np.issubdtype(classes.dtype, np.floating):
        attrs['keys_floats'] = classes
    elif np.issubdtype(classes.dtype, np.signedinteger):
        attrs['keys_int64s'] = classes
    else:
        attrs['keys_strings'] = np.array([s.encode('utf-8') for s in classes])
    attrs['values_int64s'] = np.arange(len(classes))

    cop = container.target_opset_any_domain('ai.onnx.ml')
    if cop is not None and cop < 2:
        raise RuntimeError(
            "LabelEncoder requires at least opset 2 for domain 'ai.onnx.ml' "
            "not {}".format(cop))

    container.add_node(op_type, operator.input_full_names,
                       operator.output_full_names, op_domain='ai.onnx.ml',
                       op_version=2, **attrs)


register_converter('SklearnLabelEncoder', convert_sklearn_label_encoder)



# SPDX-License-Identifier: Apache-2.0

# Modified file from
# https://github.com/onnx/onnx/blob/master/onnx/helper.py.
import numbers

from onnx import (
    TensorProto, AttributeProto,
    NodeProto, GraphProto
)
from onnx.helper import (  # noqa
    make_tensor, make_model, make_graph, _to_bytes_or_false,
    make_tensor_value_info, ValueInfoProto
)

try:
    from onnx import SparseTensorProto
    from onnx.helper import make_sparse_tensor  # noqa
except ImportError:
    # onnx is too old.
    SparseTensorProto = None

from onnx.numpy_helper import from_array  # noqa
from typing import (
    Text, Sequence, Any, Optional,
    List, cast
)
import numpy as np  # type: ignore


def make_node(
        op_type,  # type: Text
        inputs,  # type: Sequence[Text]
        outputs,  # type: Sequence[Text]
        name=None,  # type: Optional[Text]
        doc_string=None,  # type: Optional[Text]
        domain=None,  # type: Optional[Text]
        _dtype=None,  # type: [np.float32, np.float64]
        **kwargs  # type: Any
):  # type: (...) -> NodeProto
    """Construct a NodeProto.

    Arguments:
        op_type (string): The name of the operator to construct
        inputs (list of string): list of input names
        outputs (list of string): list of output names
        name (string, default None): optional unique identifier for NodeProto
        doc_string (string, default None): optional documentation
            string for NodeProto
        dtype: dtype for double used to infer
        domain (string, default None): optional domain for NodeProto.
            If it's None, we will just use default domain (which is empty)
        **kwargs (dict): the attributes of the node.  The acceptable values
            are documented in :func:`make_attribute`.
    """
    node = NodeProto()
    node.op_type = op_type
    node.input.extend(inputs)
    node.output.extend(outputs)
    if name:
        node.name = name
    if doc_string:
        node.doc_string = doc_string
    if domain is not None:
        node.domain = domain
    if kwargs:
        node.attribute.extend(
            make_attribute(key, value, dtype=_dtype, domain=domain)
            for key, value in sorted(kwargs.items()))
    return node


def make_attribute(
        key,  # type: Text
        value,  # type: Any
        dtype=None,  # type: [np.float32, np.float64]
        domain='',  # type: Text
        doc_string=None  # type: Optional[Text]
):  # type: (...) -> AttributeProto
    """Makes an AttributeProto based on the value type."""
    attr = AttributeProto()
    attr.name = key
    if doc_string:
        attr.doc_string = doc_string

    try:
        iter(value)
        is_iterable = True
    except TypeError:
        is_iterable = False
    bytes_or_false = _to_bytes_or_false(value)

    use_float64 = dtype == np.float64 and domain not in ('', 'ai.onnx.ml')

    if isinstance(value, np.float32):
        attr.f = value
        attr.type = AttributeProto.FLOAT
    elif isinstance(value, (float, np.float64)):
        if use_float64:
            attr.type = AttributeProto.TENSOR
            attr.t.CopyFrom(
                make_tensor(
                    key, TensorProto.DOUBLE, (1, ), [value]))
        else:
            attr.f = value
            attr.type = AttributeProto.FLOAT
    elif isinstance(value, np.int32):
        attr.i = value
        attr.type = AttributeProto.INT
    elif isinstance(value, np.int64):
        attr.i = value
        attr.type = AttributeProto.INT
    elif isinstance(value, numbers.Integral):
        attr.i = value
        attr.type = AttributeProto.INT
    # string
    elif bytes_or_false is not False:
        assert isinstance(bytes_or_false, bytes)
        attr.s = bytes_or_false
        attr.type = AttributeProto.STRING
    elif isinstance(value, TensorProto):
        attr.t.CopyFrom(value)
        attr.type = AttributeProto.TENSOR
    elif (SparseTensorProto is not None and
            isinstance(value, SparseTensorProto)):
        attr.sparse_tensor.CopyFrom(value)
        attr.type = AttributeProto.SPARSE_TENSOR
    elif isinstance(value, GraphProto):
        attr.g.CopyFrom(value)
        attr.type = AttributeProto.GRAPH
    # third, iterable cases
    elif is_iterable:
        byte_array = [_to_bytes_or_false(v) for v in value]
        if all(isinstance(v, np.float32) for v in value):
            attr.floats.extend(value)
            attr.type = AttributeProto.FLOATS
        elif all(isinstance(v, np.float64) for v in value):
            if use_float64:
                attr.type = AttributeProto.TENSOR
                attr.t.CopyFrom(
                    make_tensor(
                        key, TensorProto.DOUBLE, (len(value), ), value))
            else:
                attr.floats.extend(value)
                attr.type = AttributeProto.FLOATS
        elif all(isinstance(v, float) for v in value):
            if use_float64:
                attr.type = AttributeProto.TENSOR
                attr.t.CopyFrom(
                    make_tensor(
                        key, TensorProto.DOUBLE, (len(value), ), value))
            else:
                attr.floats.extend(value)
                attr.type = AttributeProto.FLOATS
        elif all(isinstance(v, np.int32) for v in value):
            attr.ints.extend(int(v) for v in value)
            attr.type = AttributeProto.INTS
        elif all(isinstance(v, np.int64) for v in value):
            attr.ints.extend(int(v) for v in value)
            attr.type = AttributeProto.INTS
        elif all(isinstance(v, numbers.Integral) for v in value):
            # Turn np.int32/64 into Python built-in int.
            attr.ints.extend(int(v) for v in value)
            attr.type = AttributeProto.INTS
        elif all(map(lambda bytes_or_false: bytes_or_false is not False,
                     byte_array)):
            attr.strings.extend(cast(List[bytes], byte_array))
            attr.type = AttributeProto.STRINGS
        elif all(isinstance(v, TensorProto) for v in value):
            attr.tensors.extend(value)
            attr.type = AttributeProto.TENSORS
        elif (SparseTensorProto is not None and
                all(isinstance(v, SparseTensorProto) for v in value)):
            attr.sparse_tensors.extend(value)
            attr.type = AttributeProto.SPARSE_TENSORS
        elif all(isinstance(v, GraphProto) for v in value):
            attr.graphs.extend(value)
            attr.type = AttributeProto.GRAPHS
        elif isinstance(value, np.matrix):
            return make_attribute(
                key, np.asarray(value).ravel(), dtype=dtype, domain=domain,
                doc_string=doc_string)
        else:
            raise ValueError(
                "You passed in an iterable attribute but the type cannot "
                "be figured out, key='{}', type={}, dtype={}, "
                "types[:5]={} set(types)={}.".format(
                    key, type(value), dtype,
                    [type(_) for _, __ in zip(value, range(0, 5))],
                    set(type(t) for t in value)))
    else:
        raise ValueError(
            "Value '{}' is not valid attribute data type for attribute "
            "'{}'.".format(value, key))
    return attr


def get_attribute_value(attr):  # type: (AttributeProto) -> Any
    if attr.type == AttributeProto.FLOAT:
        return attr.f
    elif attr.type == AttributeProto.INT:
        return attr.i
    elif attr.type == AttributeProto.STRING:
        return attr.s
    elif attr.type == AttributeProto.TENSOR:
        return attr.t
    elif attr.type == AttributeProto.GRAPH:
        return attr.g
    elif attr.type == AttributeProto.FLOATS:
        return list(attr.floats)
    elif attr.type == AttributeProto.INTS:
        return list(attr.ints)
    elif attr.type == AttributeProto.STRINGS:
        return list(attr.strings)
    elif attr.type == AttributeProto.TENSORS:
        return list(attr.tensors)
    elif attr.type == AttributeProto.GRAPHS:
        return list(attr.graphs)
    else:
        raise ValueError("Unsupported ONNX attribute: {}".format(attr))



# SPDX-License-Identifier: Apache-2.0

"""
Common functions to convert any learner based on trees.
"""
import numpy as np


def get_default_tree_classifier_attribute_pairs():
    attrs = {}
    attrs['post_transform'] = 'NONE'
    attrs['nodes_treeids'] = []
    attrs['nodes_nodeids'] = []
    attrs['nodes_featureids'] = []
    attrs['nodes_modes'] = []
    attrs['nodes_values'] = []
    attrs['nodes_truenodeids'] = []
    attrs['nodes_falsenodeids'] = []
    attrs['nodes_missing_value_tracks_true'] = []
    attrs['nodes_hitrates'] = []
    attrs['class_treeids'] = []
    attrs['class_nodeids'] = []
    attrs['class_ids'] = []
    attrs['class_weights'] = []
    return attrs


def get_default_tree_regressor_attribute_pairs():
    attrs = {}
    attrs['post_transform'] = 'NONE'
    attrs['n_targets'] = 0
    attrs['nodes_treeids'] = []
    attrs['nodes_nodeids'] = []
    attrs['nodes_featureids'] = []
    attrs['nodes_modes'] = []
    attrs['nodes_values'] = []
    attrs['nodes_truenodeids'] = []
    attrs['nodes_falsenodeids'] = []
    attrs['nodes_missing_value_tracks_true'] = []
    attrs['nodes_hitrates'] = []
    attrs['target_treeids'] = []
    attrs['target_nodeids'] = []
    attrs['target_ids'] = []
    attrs['target_weights'] = []
    return attrs


def find_switch_point(fy, nfy):
    """
    Finds the double so that
    ``(float)x != (float)(x + espilon)``.
    """
    a = np.float64(fy)
    b = np.float64(nfy)
    fa = np.float32(a)
    a0, b0 = a, a
    while a != a0 or b != b0:
        a0, b0 = a, b
        m = (a + b) / 2
        fm = np.float32(m)
        if fm == fa:
            a = m
            fa = fm
        else:
            b = m
    return a


def sklearn_threshold(dy, dtype, mode):
    """
    *scikit-learn* does not compare x to a threshold
    but (float)x to a double threshold. As we need a float
    threshold, we need a different value than the threshold
    rounded to float. For floats, it finds float *w* which
    verifies::

        (float)x <= y <=> (float)x <= w

    For doubles, it finds double *w* which verifies::

        (float)x <= y <=> x <= w
    """
    if mode == "BRANCH_LEQ":
        if dtype == np.float32:
            fy = np.float32(dy)
            if fy == dy:
                return np.float64(fy)
            if fy < dy:
                return np.float64(fy)
            eps = max(abs(fy), np.finfo(np.float32).eps) * 10
            nfy = np.nextafter([fy], [fy - eps], dtype=np.float32)[0]
            return np.float64(nfy)
        elif dtype == np.float64:
            fy = np.float32(dy)
            eps = max(abs(fy), np.finfo(np.float32).eps) * 10
            afy = np.nextafter([fy], [fy - eps], dtype=np.float32)[0]
            afy2 = find_switch_point(afy, fy)
            if fy > dy > afy2:
                return afy2
            bfy = np.nextafter([fy], [fy + eps], dtype=np.float32)[0]
            bfy2 = find_switch_point(fy, bfy)
            if fy <= dy <= bfy2:
                return bfy2
            return np.float64(fy)
        raise TypeError("Unexpected dtype {}.".format(dtype))
    raise RuntimeError("Threshold is not changed for other mode and "
                       "'BRANCH_LEQ' (actually '{}').".format(mode))


def add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id,
             feature_id, mode, value, true_child_id, false_child_id,
             weights, weight_id_bias, leaf_weights_are_counts,
             adjust_threshold_for_sklearn, dtype,
             nodes_missing_value_tracks_true=False):
    attr_pairs['nodes_treeids'].append(tree_id)
    attr_pairs['nodes_nodeids'].append(node_id)
    attr_pairs['nodes_featureids'].append(feature_id)
    attr_pairs['nodes_modes'].append(mode)
    if adjust_threshold_for_sklearn and mode != 'LEAF':
        attr_pairs['nodes_values'].append(
            sklearn_threshold(value, dtype, mode))
    else:
        attr_pairs['nodes_values'].append(value)
    attr_pairs['nodes_truenodeids'].append(true_child_id)
    attr_pairs['nodes_falsenodeids'].append(false_child_id)
    attr_pairs['nodes_missing_value_tracks_true'].append(
        nodes_missing_value_tracks_true)
    attr_pairs['nodes_hitrates'].append(1.)

    # Add leaf information for making prediction
    if mode == 'LEAF':
        flattened_weights = weights.flatten()
        factor = tree_weight
        # If the values stored at leaves are counts of possible classes, we
        # need convert them to probabilities by doing a normalization.
        if leaf_weights_are_counts:
            s = sum(flattened_weights)
            factor /= float(s) if s != 0. else 1.
        flattened_weights = [w * factor for w in flattened_weights]
        if len(flattened_weights) == 2 and is_classifier:
            flattened_weights = [flattened_weights[1]]

        # Note that attribute names for making prediction are different for
        # classifiers and regressors
        if is_classifier:
            for i, w in enumerate(flattened_weights):
                attr_pairs['class_treeids'].append(tree_id)
                attr_pairs['class_nodeids'].append(node_id)
                attr_pairs['class_ids'].append(i + weight_id_bias)
                attr_pairs['class_weights'].append(w)
        else:
            for i, w in enumerate(flattened_weights):
                attr_pairs['target_treeids'].append(tree_id)
                attr_pairs['target_nodeids'].append(node_id)
                attr_pairs['target_ids'].append(i + weight_id_bias)
                attr_pairs['target_weights'].append(w)


def add_tree_to_attribute_pairs(attr_pairs, is_classifier, tree, tree_id,
                                tree_weight, weight_id_bias,
                                leaf_weights_are_counts,
                                adjust_threshold_for_sklearn=False,
                                dtype=None):
    for i in range(tree.node_count):
        node_id = i
        weight = tree.value[i]

        if tree.children_left[i] > i or tree.children_right[i] > i:
            mode = 'BRANCH_LEQ'
            feat_id = tree.feature[i]
            threshold = tree.threshold[i]
            left_child_id = int(tree.children_left[i])
            right_child_id = int(tree.children_right[i])
        else:
            mode = 'LEAF'
            feat_id = 0
            threshold = 0.
            left_child_id = 0
            right_child_id = 0

        add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id,
                 feat_id, mode, threshold, left_child_id, right_child_id,
                 weight, weight_id_bias, leaf_weights_are_counts,
                 adjust_threshold_for_sklearn=adjust_threshold_for_sklearn,
                 dtype=dtype)


def add_tree_to_attribute_pairs_hist_gradient_boosting(
        attr_pairs, is_classifier, tree, tree_id,
        tree_weight, weight_id_bias,
        leaf_weights_are_counts,
        adjust_threshold_for_sklearn=False,
        dtype=None):
    for i, node in enumerate(tree.nodes):
        node_id = i
        weight = node['value']

        if node['is_leaf']:
            mode = 'LEAF'
            feat_id = 0
            threshold = 0.
            left_child_id = 0
            right_child_id = 0
            missing = False
        else:
            mode = 'BRANCH_LEQ'
            feat_id = node['feature_idx']
            try:
                threshold = node['threshold']
            except ValueError:
                threshold = node['num_threshold']
            left_child_id = node['left']
            right_child_id = node['right']
            missing = node['missing_go_to_left']

        add_node(attr_pairs, is_classifier, tree_id, tree_weight, node_id,
                 feat_id, mode, threshold, left_child_id, right_child_id,
                 weight, weight_id_bias, leaf_weights_are_counts,
                 adjust_threshold_for_sklearn=adjust_threshold_for_sklearn,
                 dtype=dtype, nodes_missing_value_tracks_true=missing)



# SPDX-License-Identifier: Apache-2.0

from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_sequence_at(scope: Scope, operator: Operator,
                                container: ModelComponentContainer):
    i_index = operator.index
    index_name = scope.get_unique_variable_name("seq_at%d" % i_index)
    container.add_initializer(
        index_name, onnx_proto.TensorProto.INT64, [], [i_index])
    container.add_node(
        'SequenceAt', [operator.inputs[0].full_name, index_name],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name('SequenceAt%d' % i_index))


def convert_sklearn_sequence_construct(scope: Scope, operator: Operator,
                                       container: ModelComponentContainer):
    container.add_node(
        'SequenceConstruct', [i.full_name for i in operator.inputs],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name('SequenceConstruct'))


register_converter('SklearnSequenceAt', convert_sklearn_sequence_at)
register_converter(
    'SklearnSequenceConstruct', convert_sklearn_sequence_construct)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common._apply_operation import apply_cast, apply_reshape
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_label_binariser(scope: Scope, operator: Operator,
                                    container: ModelComponentContainer):
    """Converts Scikit Label Binariser model to onnx format."""
    binariser_op = operator.raw_operator
    classes = binariser_op.classes_
    if (hasattr(binariser_op, 'sparse_input_') and
            binariser_op.sparse_input_):
        raise RuntimeError("sparse is not supported for LabelBinarizer.")
    if (hasattr(binariser_op, 'y_type_') and
            binariser_op.y_type_ == "multilabel-indicator"):
        if binariser_op.pos_label != 1:
            raise RuntimeError("pos_label != 1 is not supported "
                               "for LabelBinarizer.")
        if list(classes) != list(range(len(classes))):
            raise RuntimeError("classes != [0, 1, ..., n_classes] is not "
                               "supported for LabelBinarizer.")
        container.add_node('Identity', operator.inputs[0].full_name,
                           operator.output_full_names,
                           name=scope.get_unique_operator_name('identity'))
    else:
        zeros_tensor = np.full((1, len(classes)),
                               binariser_op.neg_label, dtype=np.float)
        unit_tensor = np.full((1, len(classes)),
                              binariser_op.pos_label, dtype=np.float)

        classes_tensor_name = scope.get_unique_variable_name('classes_tensor')
        equal_condition_tensor_name = scope.get_unique_variable_name(
            'equal_condition_tensor')
        zeros_tensor_name = scope.get_unique_variable_name('zero_tensor')
        unit_tensor_name = scope.get_unique_variable_name('unit_tensor')
        where_result_name = scope.get_unique_variable_name('where_result')

        class_dtype = onnx_proto.TensorProto.STRING
        if np.issubdtype(binariser_op.classes_.dtype, np.signedinteger):
            class_dtype = onnx_proto.TensorProto.INT64
        else:
            classes = np.array([s.encode('utf-8') for s in classes])

        container.add_initializer(classes_tensor_name, class_dtype,
                                  [len(classes)], classes)
        container.add_initializer(
            zeros_tensor_name, onnx_proto.TensorProto.FLOAT,
            zeros_tensor.shape, zeros_tensor.ravel())
        container.add_initializer(
            unit_tensor_name, onnx_proto.TensorProto.FLOAT,
            unit_tensor.shape, unit_tensor.ravel())

        reshaped_input_name = scope.get_unique_variable_name('reshaped_input')
        apply_reshape(scope, operator.inputs[0].full_name, reshaped_input_name,
                      container, desired_shape=[-1, 1])

        # Models with classes_/inputs of string type would fail in the
        # following step as Equal op does not support string comparison.
        container.add_node('Equal', [classes_tensor_name, reshaped_input_name],
                           equal_condition_tensor_name,
                           name=scope.get_unique_operator_name('equal'))
        container.add_node(
            'Where',
            [equal_condition_tensor_name, unit_tensor_name, zeros_tensor_name],
            where_result_name,
            name=scope.get_unique_operator_name('where'))
        where_res = where_result_name

        if len(binariser_op.classes_) == 2:
            array_f_name = scope.get_unique_variable_name(
                'array_feature_extractor_result')
            pos_class_index_name = scope.get_unique_variable_name(
                'pos_class_index')

            container.add_initializer(
                pos_class_index_name, onnx_proto.TensorProto.INT64, [], [1])

            container.add_node(
                'ArrayFeatureExtractor',
                [where_result_name, pos_class_index_name],
                array_f_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
            where_res = array_f_name
        apply_cast(scope, where_res, operator.output_full_names, container,
                   to=onnx_proto.TensorProto.INT64)


register_converter('SklearnLabelBinarizer', convert_sklearn_label_binariser)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common.data_types import Int64TensorType, StringTensorType
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_ordinal_encoder(scope: Scope, operator: Operator,
                                    container: ModelComponentContainer):
    ordinal_op = operator.raw_operator
    result = []
    concatenated_input_name = operator.inputs[0].full_name
    concat_result_name = scope.get_unique_variable_name('concat_result')

    if len(operator.inputs) > 1:
        concatenated_input_name = scope.get_unique_variable_name(
            'concatenated_input')
        if all(isinstance(inp.type, type(operator.inputs[0].type))
               for inp in operator.inputs):
            input_names = list(map(lambda x: x.full_name, operator.inputs))
        else:
            input_names = []
            for inp in operator.inputs:
                if isinstance(inp.type, Int64TensorType):
                    input_names.append(scope.get_unique_variable_name(
                        'cast_input'))
                    apply_cast(scope, inp.full_name, input_names[-1],
                               container, to=onnx_proto.TensorProto.STRING)
                elif isinstance(inp.type, StringTensorType):
                    input_names.append(inp.full_name)
                else:
                    raise NotImplementedError(
                        "{} input datatype not yet supported. "
                        "You may raise an issue at "
                        "https://github.com/onnx/sklearn-onnx/issues"
                        "".format(type(inp.type)))

        apply_concat(scope, input_names,
                     concatenated_input_name, container, axis=1)
    if len(ordinal_op.categories_) == 0:
        raise RuntimeError(
            "No categories found in type=%r, encoder=%r." % (
                type(ordinal_op), ordinal_op))
    for index, categories in enumerate(ordinal_op.categories_):
        attrs = {'name': scope.get_unique_operator_name('LabelEncoder')}
        if len(categories) > 0:
            if np.issubdtype(categories.dtype, np.floating):
                attrs['keys_floats'] = categories
            elif np.issubdtype(categories.dtype, np.signedinteger):
                attrs['keys_int64s'] = categories
            else:
                attrs['keys_strings'] = np.array(
                    [str(s).encode('utf-8') for s in categories])
            attrs['values_int64s'] = np.arange(
                len(categories)).astype(np.int64)

            index_name = scope.get_unique_variable_name('index')
            feature_column_name = scope.get_unique_variable_name(
                'feature_column')
            result.append(scope.get_unique_variable_name('ordinal_output'))
            label_encoder_output = scope.get_unique_variable_name(
                'label_encoder')

            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64, [], [index])

            container.add_node(
                'ArrayFeatureExtractor',
                [concatenated_input_name, index_name],
                feature_column_name, op_domain='ai.onnx.ml',
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

            container.add_node('LabelEncoder', feature_column_name,
                               label_encoder_output, op_domain='ai.onnx.ml',
                               op_version=2, **attrs)
            apply_reshape(scope, label_encoder_output, result[-1],
                          container, desired_shape=(-1, 1))
    apply_concat(scope, result, concat_result_name,
                 container, axis=1)
    cast_type = (onnx_proto.TensorProto.FLOAT if np.issubdtype(
        ordinal_op.dtype, np.floating)
        else onnx_proto.TensorProto.INT64)
    apply_cast(scope, concat_result_name, operator.output_full_names,
               container, to=cast_type)


register_converter('SklearnOrdinalEncoder', convert_sklearn_ordinal_encoder)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator


def calculate_sklearn_ovr_decision_function(operator):
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = operator.inputs[0].type.__class__(
        [N, len(operator.raw_operator.classes_)])


register_shape_calculator('SklearnOVRDecisionFunction',
                          calculate_sklearn_ovr_decision_function)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import RobustScaler, StandardScaler
from ..algebra.onnx_ops import (
    OnnxSub, OnnxDiv, OnnxCast, OnnxMul, OnnxClip, OnnxAdd,
    OnnxGather, OnnxConcat)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type, guess_proto_type
from ..proto import onnx_proto
from .common import concatenate_variables


def convert_sklearn_scaler(scope: Scope, operator: Operator,
                           container: ModelComponentContainer):
    # If there are multiple input variables, we need to combine them as a
    # whole tensor. Integer(s) would be converted to float(s).
    # Options div use true division instead of Scaler operator
    # which replaces a division by a multiplication.
    # This leads to discrepencies in some cases.
    if len(operator.inputs) > 1:
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        feature_name = operator.inputs[0].full_name
    C = operator.outputs[0].get_second_dimension()

    op = operator.raw_operator
    op_type = 'Scaler'
    attrs = {'name': scope.get_unique_operator_name(op_type)}

    if isinstance(op, StandardScaler):
        model_C = None
        if op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None and op.mean_ is not None:
            model_C = op.mean_.shape[0]
        if model_C is None and op.var_ is not None:
            model_C = op.var_.shape[0]
        if model_C is None:
            # Identity
            container.add_node(
                'Identity', feature_name,
                operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (
                    operator.outputs[0].type.shape, model_C, operator))
        C = model_C
        attrs['offset'] = (
            op.mean_ if op.with_mean else
            np.array([0.0] * C, dtype=np.float32))
        attrs['scale'] = (
            1.0 / op.scale_ if op.with_std else
            np.array([1.0] * C, dtype=np.float32))
        inv_scale = op.scale_ if op.with_std else None
    elif isinstance(op, RobustScaler):
        model_C = None
        if op.center_ is not None:
            model_C = op.center_.shape[0]
        if model_C is None and op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None:
            # Identity
            container.add_node(
                'Identity', feature_name,
                operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (
                    operator.outputs[0].type.shape, model_C, operator))
        C = model_C
        attrs['offset'] = (
            op.center_ if op.with_centering else
            np.array([0.0] * C, dtype=np.float32))
        attrs['scale'] = (
            1.0 / op.scale_ if op.with_scaling else
            np.array([1.0] * C, dtype=np.float32))
        inv_scale = op.scale_ if op.with_scaling else None
    elif isinstance(op, MaxAbsScaler):
        model_C = None
        if op.max_abs_ is not None:
            model_C = op.max_abs_.shape[0]
        if model_C is None and op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None:
            # Identity
            container.add_node(
                'Identity', feature_name,
                operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (
                    operator.outputs[0].type.shape, model_C, operator))
        C = model_C
        attrs['scale'] = 1.0 / op.scale_
        attrs['offset'] = np.array([0.] * C, dtype=np.float32)
        inv_scale = op.scale_
    else:
        raise ValueError('Only scikit-learn StandardScaler and RobustScaler '
                         'are supported but got %s. You may raise '
                         'an issue at '
                         'https://github.com/onnx/sklearn-onnx/issues.'
                         '' % type(op))

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    for k in attrs:
        v = attrs[k]
        if isinstance(v, np.ndarray) and v.dtype != dtype:
            attrs[k] = v.astype(dtype)

    use_scaler_op = container.is_allowed({'Scaler'})
    if not use_scaler_op or dtype == np.float64:
        opv = container.target_opset
        if inv_scale is None:
            sub = OnnxSub(
                feature_name, attrs['offset'].astype(dtype),
                op_version=opv,
                output_names=[operator.outputs[0].full_name])
            sub.add_to(scope, container)
        else:
            sub = OnnxSub(
                feature_name, attrs['offset'].astype(dtype),
                op_version=opv)
            div = OnnxDiv(sub, inv_scale.astype(dtype),
                          op_version=opv,
                          output_names=[operator.outputs[0].full_name])
            div.add_to(scope, container)
        return

    if inv_scale is not None:
        options = container.get_options(op, dict(div='std'))
        div = options['div']
        if div == 'div':
            opv = container.target_opset
            sub = OnnxSub(
                feature_name, attrs['offset'].astype(dtype),
                op_version=opv)
            div = OnnxDiv(sub, inv_scale.astype(dtype),
                          op_version=opv,
                          output_names=[operator.outputs[0].full_name])
            div.add_to(scope, container)
            return
        if div == 'div_cast':
            opv = container.target_opset
            cast = OnnxCast(feature_name, to=onnx_proto.TensorProto.DOUBLE,
                            op_version=opv)
            sub = OnnxSub(cast, attrs['offset'].astype(np.float64),
                          op_version=opv)
            div = OnnxDiv(sub, inv_scale.astype(np.float64), op_version=opv)
            cast = OnnxCast(div, to=proto_dtype, op_version=opv,
                            output_names=[operator.outputs[0].full_name])
            cast.add_to(scope, container)
            return

    if attrs['offset'].size != attrs['scale'].size:
        # Scaler does not accept different size for offset and scale.
        size = max(attrs['offset'].size, attrs['scale'].size)
        ones = np.ones(size, dtype=attrs['offset'].dtype)
        attrs['offset'] = attrs['offset'] * ones
        attrs['scale'] = attrs['scale'] * ones

    container.add_node(
        op_type, feature_name, operator.outputs[0].full_name,
        op_domain='ai.onnx.ml', **attrs)


def convert_sklearn_min_max_scaler(
        scope: Scope, operator: Operator,
        container: ModelComponentContainer):
    # If there are multiple input variables, we need to combine them as a
    # whole tensor. Integer(s) would be converted to float(s).
    # Options div use true division instead of Scaler operator
    # which replaces a division by a multiplication.
    # This leads to discrepencies in some cases.
    if len(operator.inputs) > 1:
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        feature_name = operator.inputs[0].full_name

    op = operator.raw_operator
    opv = container.target_opset

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    # X *= self.scale_
    # X += self.min_
    # if self.clip:
    #     np.clip(X, self.feature_range[0], self.feature_range[1], out=X)
    casted = OnnxCast(feature_name, to=proto_dtype, op_version=opv)
    scaled = OnnxMul(casted, op.scale_.astype(dtype),
                     op_version=opv)

    if getattr(op, 'clip', False):
        # parameter clip was introduced in scikit-learn 0.24
        offset = OnnxAdd(scaled, op.min_.astype(dtype),
                         op_version=opv)
        collect = []
        for i in range(op.data_min_.shape[0]):
            gather = OnnxGather(offset, np.array([i], dtype=np.int64),
                                axis=1, op_version=opv)
            clipped = OnnxClip(gather, op.data_min_[i:i+1].astype(dtype),
                               op.data_max_[i:i+1].astype(dtype),
                               op_version=opv)
            collect.append(clipped)
        concat = OnnxConcat(*collect, op_version=opv, axis=1,
                            output_names=[operator.outputs[0].full_name])
        concat.add_to(scope, container)
    else:
        offset = OnnxAdd(scaled, op.min_.astype(dtype),
                         op_version=opv,
                         output_names=[operator.outputs[0].full_name])
        offset.add_to(scope, container)


register_converter('SklearnRobustScaler', convert_sklearn_scaler,
                   options={'div': ['std', 'div', 'div_cast']})
register_converter('SklearnScaler', convert_sklearn_scaler,
                   options={'div': ['std', 'div', 'div_cast']})
register_converter('SklearnMinMaxScaler', convert_sklearn_min_max_scaler)
register_converter('SklearnMaxAbsScaler', convert_sklearn_scaler,
                   options={'div': ['std', 'div', 'div_cast']})
